var isFirefox = navigator.userAgent.indexOf("Firefox") > -1;
var pixelRatio;
var captureImages = [];

function getChromeVersion () {     
    var raw = navigator.userAgent.match(/Chrom(e|ium)\/([0-9]+)\./);
    return raw ? parseInt(raw[2], 10) : false;
}

window.jQuery = $;
var chromeVersion = getChromeVersion() || 0;

var MAX_PRIMARY_DIMENSION = 32750,
    MAX_AREA = MAX_PRIMARY_DIMENSION * MAX_PRIMARY_DIMENSION;

if (chromeVersion >= 76) {
    MAX_PRIMARY_DIMENSION = 65500;
}

var rect = undefined;

var matches = ['http://*/*', 'https://*/*', 'ftp://*/*', 'file://*/*'],
    noMatches = [/^https?:\/\/chrome.google.com\/.*$/];

function isValidUrl(url) {
    var r, i;
    for (i = noMatches.length - 1; i >= 0; i--) {
        if (noMatches[i].test(url)) {
            return false;
        }
    }
    for (i = matches.length - 1; i >= 0; i--) {
        r = new RegExp('^' + matches[i].replace(/\*/g, '.*') + '$');
        if (r.test(url)) {
            return true;
        }
    }
    return false;
}


function initiateCapture(tab) {
    captureImages = [];
    chrome.tabs.sendMessage(tab.id, { msg: 'scrollPage' }, function() {});
}

function captureVisibleTab (data, sendResponse) {
    var format = localStorage.getItem("captureFormat") || 'png';
    chrome.tabs.captureVisibleTab(null, {format: format, quality: 100}, function(dataURI) {
        if (dataURI) {
            var image = new Image();
            var onload = function () {
                image.removeEventListener("load", onload);

                captureImages.push({
                    image: this,
                    data: data
                });

                sendResponse(true);
                if (data.complete === 1) {
                    generateCaptureCanvas();
                }
            };
            image.addEventListener("load", onload);
            image.src = dataURI;
        }
    });
}

function generateCaptureCanvas() {

    // 計算需要创建多大的 canvas
    var startY = 99999999;
    var captureHiehgt = captureImages[0].image.height;

    for (var i = 0; i < captureImages.length; i++) {
        var img = captureImages[i];
        var y = (img.data.y * pixelRatio);
        if (y < startY) {
            startY = y;
        }
    }

    // 计算需要多少张 Canvas，并把对应的截图结果保存下来
    var canvasImageLists = [];
    var tempList = [];
    var currentY = 0;
    captureImages = captureImages.sort(function (a, b) {
        return a.data.y - b.data.y;
    });
    for (var i = 0; i < captureImages.length; i++) {
        var captureImage = captureImages[i];
        var imageY = (captureImage.data.y * pixelRatio) - startY - currentY + MAX_PRIMARY_DIMENSION * canvasImageLists.length;
        if (imageY + captureImage.image.height > MAX_PRIMARY_DIMENSION * (canvasImageLists.length + 1)  - captureImage.image.height ) {
            currentY = imageY;
            canvasImageLists.push(tempList);
            sum = 0;
            tempList = [];
        }
        tempList.push(captureImage);
    }
    if (tempList.length > 0) {
        canvasImageLists.push(tempList);
    }

    // 开始拼接截图
    var resultCanvas = [];
    var currentY = 0;
    canvasImageLists.forEach(function (canvasImages) {

        // 计算 canvas 高度
        var canvasHeight = 0;
        canvasImages = canvasImages.sort(function (a, b) {
            return b.data.y - a.data.y;
        });

        canvasHeight = (canvasImages[0].data.y * pixelRatio - startY) + captureHiehgt - currentY;

        var canvas = document.createElement('canvas');
        var ctx = canvas.getContext('2d');
        canvas.height = Math.min(canvasHeight, MAX_PRIMARY_DIMENSION);
        canvas.width = canvasImages[0].image.width;
        canvasImages.forEach(function (img, index) {
            var y = (img.data.y * pixelRatio) - startY - currentY;
            ctx.drawImage(img.image, 0, y);
            img.image.src = "";
            img.image = null;
            delete img.image;
        });

        // 裁切模式
        if (rect) {
            var tempCanvas = document.createElement('canvas');
            var tctx = tempCanvas.getContext('2d');
            tempCanvas.width = (rect.width * pixelRatio);
            tempCanvas.height = (rect.height * pixelRatio);
            var x = (rect.x * pixelRatio);
            var y = (rect.y * pixelRatio) - startY;
            var width = (rect.width * pixelRatio);
            var height = (rect.height * pixelRatio);
            var imgData = ctx.getImageData(x, y, width, height);
            tctx.putImageData(imgData, 0, 0);
            if (pixelRatio > 1 && !useRetina) {
                resultCanvas.push(resizeCanvas(tempCanvas, pixelRatio));
            }
            else {
                resultCanvas.push(tempCanvas);
            }
        }
        // 整页截图
        else {
            if (pixelRatio > 1 && !useRetina) {
                resultCanvas.push(resizeCanvas(canvas, pixelRatio));
            }
            else {
                resultCanvas.push(canvas);
            }
        }

        currentY += canvasHeight;
    });
    captureImages = [];
    returnCanvases(resultCanvas);
    resultCanvas.forEach(function (canvas) {
        var ctx = canvas.getContext('2d');
        ctx.clearRect(0, 0, canvas.width, canvas.height);
        canvas = null;
    });
}

function resizeCanvas(originalCanvas, pixelRatio) {
    var canvas = document.createElement('canvas');
    var ctx = canvas.getContext("2d");
    var cw = originalCanvas.width * 1/pixelRatio;
    var ch = originalCanvas.height * 1/pixelRatio;

    canvas.width = cw;
    canvas.height = ch;

    ctx.drawImage(originalCanvas, 0, 0, cw, ch);
    return canvas;
}

function startCapture (tab) {

    var loaded = false,
        screenshots = [],
        timeout = 3000,
        timedOut = false,
        noop = function() {};

    if (!isValidUrl(tab.url)) {
        errback('invalid url');
    }

    rect = undefined;

    chrome.tabs.executeScript(tab.id, {file: 'page.js'}, function() {
        if (timedOut) {
            console.error('Timed out too early while waiting for ' +
                          'chrome.tabs.executeScript. Try increasing the timeout.');
        } 
        else {
            loaded = true;
            // 开始滚屏
            initiateCapture(tab);
        }
    });

    window.setTimeout(function() {
        if (!loaded) {
            timedOut = true;
            errback('execute timeout');
        }
    }, timeout);
}

chrome.runtime.onMessage.addListener(function (request, sender, sendResponse) {
    if (request.msg === 'capture') {
        rect = request.rect || undefined;
        captureVisibleTab(request, sendResponse);
    }
    return true;
});

function webCapture(tab, filename, callback, errback, devicePixelRatio) {

    pixelRatio = devicePixelRatio || 1;
    if (isFirefox) pixelRatio = 1;
    
    startCapture(tab);
}

function returnCanvases (canvases) {
    var format = localStorage.getItem("captureFormat") || 'png';
    if (canvases.length > 15) {
        canvas.length = 15
    }
    canvases.forEach(function (canvas, index) {
        if (canvas) {
            var base64String;
            if (format === "png") {
                base64String = canvas.toDataURL();
            }
            else {
                base64String = canvas.toDataURL("image/jpeg", 1);
            }
            (function(b64, index) {
                setTimeout(function () {
                    if (canvases.length > 1) {
                        finishCaptures(b64, index + 1);
                    }
                    else {
                        finishCaptures(b64);
                    }
                }, index * 500);
            })(base64String, index);
        }
    });
}